/*
 * allocmem.c:
 *	allocmem is a simple memory allocator that replaces PsyQ's malloc.
 *	The actual blocks of memory are stored in a buffer called "allocbuf,"
 *	and each allocation is tracked via a 2D array named "alloctable."
 */

#include <engine/psx.h>
#include <engine/allocmem.h>

#define ALLOCBUF_SIZE		0x3a000
#define ALLOCBUF_DC_SIZE	0x3ff
#define MAX_ALLOC		500
#define MAX_ALLOC_DC		10

/*  
 *  This is how allocbuf should look upon program start:
 *  
 *  ________________________________________________ 
 * |                                                |
 * |                 ALLOCBUF_SIZE                  |
 * |________________________________________________|
 *  <---------------------free---------------------> 
 */
static char allocbuf[ALLOCBUF_SIZE];
static char *allocbuf_dc = (char *)0x1f800000;
static char *p_allocbuf = NULL;
static unsigned int allocbuf_size_max = 0;

/*
 *                   EXAMPLE alloctable
 *                   ~~~~~~~~~~~~~~~~~~
 *
 *                           0      1                   n-2    n-1   
 *                         _________________________________________ 
 *                        |      |      |      |      |      |      |
 *   MEMORY ADDRESS     0 | 0x03 | 0x80 |             | NULL | NULL |
 *                        |______|______|____ ... ____|______|______|
 *                        |      |      |             |      |      |
 *        SIZE          1 |  125 |  400 |             | NULL | NULL |
 *                        |______|______|______|______|______|______|
 */
static unsigned int alloctable[2][MAX_ALLOC];
static unsigned int alloctable_dc[2][MAX_ALLOC_DC];
static unsigned int *p_alloctable = NULL;
static unsigned int alloctable_max = 0;

/*
 * determine_palloc:
 *	Sets pointers for allocation buffer and allocation table.
 */
static
determine_palloc(bufidx)
unsigned int bufidx;
{
	switch(bufidx)
	{
		case ALLOCBUF:
			p_allocbuf = &allocbuf[0];
			p_alloctable = alloctable;
			allocbuf_size_max = ALLOCBUF_SIZE;
			alloctable_max = MAX_ALLOC;
			break;
		case ALLOCBUF_DC:
			p_allocbuf = allocbuf_dc;
			p_alloctable = alloctable_dc;
			allocbuf_size_max = ALLOCBUF_DC_SIZE;
			alloctable_max = MAX_ALLOC_DC;
			break;
		default:
			printf("determine_palloc() error: invalid buffer index! (%d)\n", bufidx);
			return 0;
	}
	return 1;
}


/* FIXME: RENAME WITH ALLOCMEM_ PREFIX
 * determine_allocs_used:
 *	Returns the number of memory allocations the program has made.
 */
unsigned int
determine_allocs_used(bufidx)
unsigned int bufidx;
{
	int i, allocs_used = 0;

	if(!determine_palloc(bufidx))
	{
		return;
	}

	for(i = 0; i < alloctable_max; i++)
	{
		if(*(p_alloctable+i) == (unsigned int)NULL)
		{
			return allocs_used;
		}
		allocs_used++;
	}
	return allocs_used;
}


/*
 * allocmem_clear:
 *	Clears both allocbuf and alloctable, effectively resetting the
 *	whole system. Should be called during initialization.
 */
void
allocmem_clear(bufidx)
unsigned int bufidx;
{
	int row, col;

	if(!determine_palloc(bufidx))
	{
		return;
	}

	/* set all elements of allocbuf to zero */
	for(col = 0; col < allocbuf_size_max; col++)
	{
		p_allocbuf[col] = 0;
	}

	/* set all elements of alloctable to NULL */
	for(row = 0; row < 2; row++)
	{
		for(col = 0; col < alloctable_max; col++)
		{
			*(p_alloctable+(row*alloctable_max)+col) = 0;
		}
	}
}

/*
 * allocmem_space_used:
 *	Prints the amount of space currently in use by allocmem.
 */
unsigned int
allocmem_space_used(bufidx)
{
	unsigned int col, space_used = 0;

	if(!determine_palloc(bufidx))
	{
		return;
	}

	for(col = 0; col < alloctable_max; col++)
	{
		space_used += *(p_alloctable+alloctable_max+col);
	}

	return space_used;
}

/*
 * allocmem_print:
 *	Prints contents of alloctable onto stdout. Useful for debugging.
 */
void
allocmem_print(bufidx)
unsigned int bufidx;
{
	unsigned int col, space_used = 0;

	if(!determine_palloc(bufidx))
	{
		return;
	}

	for(col = 0; col < alloctable_max; col++)
	{
		printf("[%u] 0x%x : %u\n", col, *(p_alloctable+col), *(p_alloctable+alloctable_max+col));
	}

	printf("\nspace used: %u/%u bytes\n", allocmem_space_used(bufidx), allocbuf_size_max);
}

/*
 * allocmem:
 *	Attempts to allocate a block of memory in allocbuf, then record it
 *	in alloctable. On success it returns the memory address of the newly
 *	created block. On failure it returns NULL.
 */
void
*allocmem(bufidx, nobj, size)
unsigned int bufidx, nobj, size;
{
	/*
	 * nobj -- number of objects
	 * size -- size of each object
	 */

	int i, j;
	int allocs_used = 0;

	if(!determine_palloc(bufidx))
	{
		return;
	}

	allocs_used = determine_allocs_used();

	/* zero-allocations should not take up space in the alloctable! */
	if(!nobj || !size)
	{
		allocmem_error = 1;
		return;
	}

	for(i = 0; i < alloctable_max; i++)
	{
		/*
		 * if the slot is NULL, then we are to the right of all
		 * allocated pieces of memory specified by alloctable.
		 */
		if(*(p_alloctable+i) == (unsigned int)NULL)
		{
			/* the first allocation always starts at allocbuf */
			if(i == 0)
			{
				/*
				 * make sure that the requested size does not exceed the
				 * upper bound of allocbuf
				 */
				if(nobj*size > allocbuf_size_max)
				{
					allocmem_error = 2;
					return NULL;
				}

				*(p_alloctable+i) = (unsigned int)p_allocbuf;
				*(p_alloctable+alloctable_max+i) = nobj*size;

				printf("===========0x%x===========\n", *(p_alloctable+i));

				return (void *)(*(p_alloctable+i));
			}

			/*
			 * take the end position of the furthest allocated piece
			 * of memory, add the requested size from there, and see if
			 * it ends before the upper bound of allocbuf. if that is true,
			 * use that as the start position of the requested allocated
			 * piece of memory. if not, return NULL.
			 */
			else if( ((*(p_alloctable+i-1)) + (*(p_alloctable+alloctable_max+i-1)) + nobj*size) < (unsigned int)p_allocbuf + allocbuf_size_max)
			{
				/* append data to the end of the table */
				*(p_alloctable+i) = (*(p_alloctable+i-1)) + (*(p_alloctable+alloctable_max+i-1));
				*(p_alloctable+alloctable_max+i) = nobj*size;

				return (void *)(*(p_alloctable+i));
			}
			else
			{
				allocmem_error = 3;
				printf("\nallocbuf_max: %d \nallocbuf range(0x%x -- 0x%x)\n", allocbuf_size_max, &allocbuf[0], &allocbuf[ALLOCBUF_SIZE-1]);
				printf("0x%x\n0x%x\nnobj*size: %d\nend: 0x%x\n", (*(p_alloctable+i-1)), (*(p_alloctable+alloctable_max+i-1)), nobj*size, (unsigned int)p_allocbuf + allocbuf_size_max);
				printf("allocbuf_start: 0x%x\n", allocbuf);
				return NULL;
			}
		}

		/*
		 * we are NOT at the end of the table. since we are in the middle of the table,
		 * try to find free space between pieces of already allocated memory. if the
		 * requested size fits in between any of these spaces, ensure that all memory
		 * addresses are IN ORDER from least to greatest within the table itself. this
		 * is done by shifting elements to the right.
		 */

		else
		{
			/* find a possible space for the new allocation */
			if(allocs_used < alloctable_max)
			{
				/*
				 * check if there's enough memory before the start address of the zeroth allocation.
				 * this is particularly useful if you end up allocating an initial large chunk of memory
				 * and then free it after subsequent allocations.
				 */
				if( !i && ((*p_alloctable)-((unsigned int)p_allocbuf) > nobj*size) )
				{
					for(j = allocs_used-1; j >= 0; j--)
					{
						*(p_alloctable+j+1) = *(p_alloctable+j);
						*(p_alloctable+alloctable_max+j+1) = *(p_alloctable+alloctable_max+j);
					}


					*p_alloctable = p_allocbuf;
					*(p_alloctable+alloctable_max) = nobj*size;
					return (void *)(*p_alloctable);
				}

				/*
				 * use the next slot because it's NULL! it doesn't house a piece of
				 * already allocated memory, so see if we can use it.
				 */
				else if(*(p_alloctable+i+1) == (unsigned int)NULL)
				{
					continue;
				}

				/*
				 * otherwise, the next slot contains a piece of already allocated memory,
				 * and we must see if there is enough free space between the two allocations
				 * that will fit the size of our requested allocation.
				 */
				else if( ((*(p_alloctable+i+1)) - ((*(p_alloctable+i)) + (*(p_alloctable+alloctable_max+i)))) >= nobj*size )
				{
					/*
					 * create new slot for the allocation by shifting everything after
					 * slot "i" forward starting at the end of alloctable going
					 * backwards.
					 */
					for(j = allocs_used-1; j > i; j--)
					{
						*(p_alloctable+j+1) = *(p_alloctable+j);
						*(p_alloctable+alloctable_max+j+1) = *(p_alloctable+alloctable_max+j);
					}

					/* insert the data! */
					*(p_alloctable+i+1) = (*(p_alloctable+i)) + (*(p_alloctable+alloctable_max+i));
					*(p_alloctable+alloctable_max+i+1) = nobj*size;

					//allocmem_print(ALLOCBUF);
					//printf(" (%d-(%d+%d)) >= %d\n ", (*p_alloctable+i+1), ((*p_alloctable+i), (*(p_alloctable+alloctable_max+i)), nobj*size));
					//printf("%d vs %d\n", (*p_alloctable+i+1), *(p_alloctable+i+1) );
					//while(1);

					/* return a pointer to the address */
					return (void *)(*(p_alloctable+i+1));
				}
			}
			/* all allocation pieces have been used! */
			else
			{
				allocmem_error = 4;
				return NULL;
			}
		}
	}

	/*
	 * we have ultimately failed to find any free space for the requested
	 * allocation, so return NULL.
	 */

	allocmem_error = 5;
	return NULL;
}

/*
 * allocmem_align:
 *	Aligns allocbuf by adding another allocation if needed.
 */
void
allocmem_align(bufidx)
unsigned int bufidx;
{
	unsigned int align;
	int allocidx = determine_allocs_used(bufidx)-1;

	if(!determine_palloc(bufidx) || allocidx < 0)
	{
		return;
	}

	//printf("%d\n%d\n", p_alloctable[0][allocidx], p_alloctable[1][allocidx]);

	align = (*(p_alloctable+allocidx)) + (*(p_alloctable+alloctable_max+allocidx));
	if(align%4)
	{
		allocmem(bufidx, 1, align);
	}
}

/*
 * allocmem_free:
 *	Attempts to free a block of memory from allocbuf, then remove its
 *	record from alloctable.
 */
void
allocmem_free(bufidx, p_addr)
unsigned int bufidx;
void *p_addr;
{
	int i, j;
	int allocs_used;

	if(!determine_palloc(bufidx))
	{
		return;
	}

	allocs_used = determine_allocs_used();

	for(i = 0; i < alloctable_max; i++)
	{
		if(*(p_alloctable+i) == (unsigned int)p_addr)
		{
			/*
			 * if the element is at the end of the table, then set it to zero.
			 * no shifting required.
			 */
			if(i == allocs_used-1)
			{
				*(p_alloctable+i) = 0;
				*(p_alloctable+alloctable_max+i) = 0;

				return;
			}

			/*
			 * if the element to be freed is before the end of the table, we must
			 * shift elements to the left starting at slot "i" going forward.
			 */
			else if(i < allocs_used-1)
			{
				for(j = i+1; j < allocs_used; j++)
				{
					*(p_alloctable+j-1) = *(p_alloctable+j);
					*(p_alloctable+alloctable_max+j-1) = *(p_alloctable+alloctable_max+j);

					*(p_alloctable+j) = 0;
					*(p_alloctable+alloctable_max+j) = 0;
				}

				return;
			}
		}
	}
}

